import sympy as sp
import random
import copy
from typing import List
from pystencils import Field, Assignment

fa_symbol_iter = sp.numbered_symbols("fa_")


def merge_field_accesses(assignments):
    """Transformation that introduces symbols for all read field accesses
    for multiple read accesses only one symbol is introduced"""
    field_accesses = {}

    new_eqs = copy.copy(assignments)
    for assignment in new_eqs:
        for arg in assignment.rhs.atoms():
            if isinstance(arg, Field.Access) and arg not in field_accesses:
                field_accesses[arg] = next(fa_symbol_iter)

    for i in range(0, len(new_eqs)):
        for f, s in field_accesses.items():
            if f in new_eqs[i].atoms():
                new_eqs[i] = new_eqs[i].subs(f, s)

    for f, s in field_accesses.items():
        new_eqs.insert(0, Assignment(s, f))

    return new_eqs


def fuse_eqs(input_eqs, max_depth=1, max_usage=1):
    """Inserts subexpressions that are used not more than `max_usage`

    Args:
        max_depth: complexity metric for the subexpression to insert
                   if max_depth is larger than the expression tree of the subexpression
                   the subexpressions is not inserted

    Somewhat the inverse of common subexpression elimination.
    """
    eqs = copy.copy(input_eqs)
    usages = get_usage(eqs)
    definitions = get_definitions(eqs)

    def inline_trivially_schedulable(sym, depth):

        if sym not in definitions or sym not in usages or usages[sym] > max_usage or depth > max_depth:
            return sym

        rhs = definitions[sym].rhs
        if len(rhs.args) == 0:
            return rhs

        return rhs.func(*[inline_trivially_schedulable(arg, depth + 1) for arg in rhs.args])

    for idx, eq in enumerate(eqs):
        if usages[eq.lhs] > 1 or isinstance(eq.lhs, Field.Access):
            if not isinstance(eq.rhs, sp.Symbol):
                eqs[idx] = Assignment(
                    eq.lhs,
                    eq.rhs.func(*[inline_trivially_schedulable(arg, 0) for arg in eq.rhs.args]))

    count = 0
    while len(eqs) != count:
        count = len(eqs)
        usages = get_usage(eqs)
        eqs = [eq for eq in eqs if usages[eq.lhs] > 0 or isinstance(eq.lhs, Field.Access)]

    return eqs


def schedule_eqs(assignments: List[Assignment], candidate_count=20):
    """Changes order of assignments to save registers.

    Args:
        assignments:
        candidate_count: tuning parameter, small means fast, but bad scheduling quality
                         1 corresponds to full greedy search

    Returns:
        list of re-ordered assignments
    """
    if candidate_count == 0:
        return assignments

    definitions = get_definitions(assignments)
    definition_atoms = {}
    for sym, definition in definitions.items():
        definition_atoms[sym] = list(definition.rhs.atoms(sp.Symbol))
    roots = get_roots(assignments)
    initial_usages = get_usage(assignments)

    level = 0
    current_level_set = set([frozenset(roots)])
    current_usages = {frozenset(roots): {u: 0 for u in roots}}
    current_schedules = {frozenset(roots): (0, [])}
    max_regs = 0
    while len(current_level_set) > 0:
        new_usages = dict()
        new_schedules = dict()
        new_level_set = set()

        min_regs = min([len(current_usages[dec_set]) for dec_set in current_level_set])
        max_regs = max(max_regs, min_regs)

        def score_dec_set(dec_set):
            score = len(current_usages[dec_set])  # current_schedules[dec_set][0]
            return dec_set, score

        candidates = [score_dec_set(dec_set) for dec_set in current_level_set]

        random.shuffle(candidates)
        candidates.sort(key=lambda d: d[1])

        for dec_set, regs in candidates[:candidate_count]:

            for dec in dec_set:
                new_dec_set = set(dec_set)
                new_dec_set.remove(dec)
                usage = dict(current_usages[dec_set])
                usage.pop(dec)
                atoms = definition_atoms[dec]
                for arg in atoms:
                    if not isinstance(arg, Field.Access):
                        argu = usage.get(arg, initial_usages[arg]) - 1
                        if argu == 0 and arg in definitions:
                            new_dec_set.add(arg)
                        usage[arg] = argu
                frozen_new_dec_set = frozenset(new_dec_set)
                schedule = current_schedules[dec_set]
                max_reg_count = max(len(usage), schedule[0])

                if frozen_new_dec_set not in new_schedules or max_reg_count < new_schedules[frozen_new_dec_set][0]:
                    new_schedule = list(schedule[1])
                    new_schedule.append(definitions[dec])
                    new_schedules[frozen_new_dec_set] = (max_reg_count, new_schedule)

                if len(frozen_new_dec_set) > 0:
                    new_level_set.add(frozen_new_dec_set)
                new_usages[frozen_new_dec_set] = usage

        current_schedules = new_schedules
        current_usages = new_usages
        current_level_set = new_level_set

        level += 1

    schedule = current_schedules[frozenset()]

    schedule[1].reverse()
    return schedule[1]


def liveness_opt_transformation(eqs):
    return fuse_eqs(merge_field_accesses(schedule_eqs(eqs, 30)), 1, 3)


# ---------- Utilities   -----------------------------------------------------------------------------------------


def get_usage(assignments: List[Assignment]):
    """Count number of reads for all symbols in list of assignments

    Returns:
        dictionary mapping symbol to number of its reads
    """
    reg_usage = {}
    for assignment in assignments:
        for arg in assignment.rhs.atoms():
            if isinstance(arg, sp.Symbol) and not isinstance(arg, Field.Access):
                if arg in reg_usage:
                    reg_usage[arg] += 1
                else:
                    reg_usage[arg] = 1
    return reg_usage


def get_definitions(assignments: List[Assignment]):
    """Returns dictionary mapping symbol to its defining assignment"""
    definitions = {}
    for assignment in assignments:
        definitions[assignment.lhs] = assignment
    return definitions


def get_roots(eqs):
    """Returns all field accesses that are used as lhs in assignment (stores)
    In case there are no independent assignments, the last one is returned (TODO try if necessary)
    """
    roots = []
    for eq in eqs:
        if isinstance(eq.lhs, Field.Access):
            roots.append(eq.lhs)
    if not roots:
        roots.append(eqs[-1].lhs)
    return roots


# ---------- Staggered kernels -----------------------------------------------------------------------------------------

def unpack_staggered_eqs(field, expressions, subexpressions):
    eqs = copy.deepcopy(subexpressions)
    for dim in range(0, len(expressions)):
        for vec in range(0, len(expressions[dim])):
            eqs.append(Assignment(Field.Access(field, (0, 0, 0, dim, vec)), expressions[dim][vec]))
    return eqs


def pack_staggered_eqs(eqs, field, expressions, subexpressions):
    new_matrix_list = [0] * (field.shape[-1] * field.shape[-2])

    for eq in eqs:
        if isinstance(eq.lhs, Field.Access):
            new_matrix_list[eq.lhs.offsets[-2] * field.shape[-1] + eq.lhs.offsets[-1]] = eq.rhs

    subexpressions = [eq for eq in eqs if not isinstance(eq.lhs, Field.Access)]

    return (field, [
        sp.Matrix(field.shape[-1], 1,
               new_matrix_list[dim * field.shape[-1]:(dim + 1) * field.shape[-1]])
        for dim in range(field.shape[-2])
    ], subexpressions)
