# coding: utf-8

# In[32]:

import warnings
import pystencils as ps
from pygrandchem.grandchem import GrandChemGenerator
from pygrandchem.scenarios import system_4_2, system_3_1
from pygrandchem.initialization import init_boxes, smooth_fields
from pygrandchem.scenarios import benchmark_configs

from sympy import Number, Symbol, Expr, preorder_traversal, postorder_traversal, Function, Piecewise, relational
from pystencils.simp import sympy_cse_on_assignment_list
from pystencils.simp.liveness_opts import *
from pystencils.simp.liveness_opts_exp import *

import graphviz

import pycuda

import sys
from subprocess import run, PIPE

from pystencils import show_code
import pycuda.driver as drv

configs = benchmark_configs()

def get_config(name):
    return configs[name]


domain_size = (512, 512, 128)
periodicity = (True, True, False)

optimization = {'gpu_indexing_params': {"block_size": (32, 4, 2)}}
config = get_config('42_fixT')
phases, components = config['Parameters']['phases'], config['Parameters']['components']
format_args = {'p': phases, 'c': components, 's': ','.join(str(e) for e in domain_size)}

# Adding fields
dh = ps.create_data_handling(domain_size, periodicity=periodicity, default_target='gpu')
f = dh.fields
phi_src = dh.add_array(
    'phi_src', values_per_cell=config['Parameters']['phases'], layout='fzyx', latex_name='phi_s')
mu_src = dh.add_array(
    'mu_src', values_per_cell=config['Parameters']['components'], layout='fzyx', latex_name="mu_s")
mu_stag = dh.add_array(
    'mu_stag', values_per_cell=(dh.dim, config['Parameters']['components']), layout='f')
phi_stag = dh.add_array('phi_stag', values_per_cell=(dh.dim, phases), layout='f')

phi_dst = dh.add_array_like('phi_dst', 'phi_src')
mu_dst = dh.add_array_like('mu_dst', 'mu_src')

gc = GrandChemGenerator(
    phi_src,
    phi_dst,
    mu_src,
    mu_dst,
    config['FreeEnergy'],
    config['Parameters'],
    #conc=c,
    mu_staggered=mu_stag,
    phi_staggered=phi_stag,
    use_block_offsets=False,
    compile_kernel=False)

mu_full_eqs = gc.mu_full()
phi_full_eqs = gc.phi_full()

mu_partial1_eqs = gc.mu_partial1()
mu_partial2_eqs = gc.mu_partial2()

phi_partial1_eqs = gc.phi_partial1()
phi_partial2_eqs = gc.phi_partial2()

phi_kernel = ps.create_kernel(phi_full_eqs, target='gpu', **optimization).compile()
mu_kernel = ps.create_kernel(mu_full_eqs, target='gpu', **optimization).compile()

c = dh.add_array('c', values_per_cell=config['Parameters']['components'], layout='fzyx', gpu=False)

init_boxes(dh)
#initialize_concentration_field(dh, free_energy, config['Parameters']['initial_concentration'])
smooth_fields(dh, sigma=0.4, iterations=5, dim=dh.dim)
dh.synchronization_function(['phi_src', 'phi_dst', 'mu_src', 'mu_dst'])()
print(dh)


def bench_kernels(mu_kernel, phi_kernel):

    start = drv.Event()
    end = drv.Event()

    dh.run_kernel(mu_kernel)
    start.record()
    dh.run_kernel(mu_kernel)
    dh.run_kernel(mu_kernel)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    print("mu_kernel: {}  {:5.3f} ms".format(mu_kernel.num_regs, msec))

    dh.run_kernel(phi_kernel)
    start.record()
    dh.run_kernel(phi_kernel)
    dh.run_kernel(phi_kernel)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    print("phi_kernel: {}  {:5.3f} ms".format(phi_kernel.num_regs, msec))


sched_options = []

sched_options.append(option_none)
#sched_options.append(option_reschedule)
sched_options.append(option_liveness_opt_transformation)
#sched_options.append(option_dupl_reschedule)
sched_options.append(option_liveness_opt_transformation_shmem)
sched_options.append(option_liveness_opt_transformation_shmem2)
#sched_options.append(option_reschedule_shmem)
#sched_options.append(option_liveness_opt_transformation_shmem)
#sched_options.append(optionFuseSubs)
#sched_options.append(optionFuseFMAs)
#sched_options.append(optionFuseBoth)
#sched_options.append(optionRescheduleAtomize)
#sched_options.append(optionRescheduleAtomizeScramble)
#sched_options.append(optionDuplAtomizeReschedule)
#sched_options.append(optionDuplAtomizeRefuseReschedule)
#sched_options.append(optionDuplRescheduleAtomize)
#sched_options.append(optionDuplRescheduleAtomizeScramble)
#sched_options.append(optionSchedIteration)
#sched_options.append(optionAtomizeRescheduleNoSqrt)
#sched_options.append(optionAtomizeRescheduleNoDiv)
#sched_options.append(optionAtomizeRescheduleNoSqrtDiv)
#sched_options.append(optionAtomizeRescheduleNoPiecewise)
#sched_options.append(optionAtomizeRescheduleNoAll)

print("warmup")
bench_kernels(mu_kernel, phi_kernel)
dh.swap('mu_src', 'mu_dst')
dh.swap('phi_src', 'phi_dst')

for sched_option in sched_options:
    mu_full_opt_eqs = sched_option(mu_full_eqs)
    phi_full_opt_eqs = sched_option(phi_full_eqs)


    #mu_partial1_opt_eqs = sched_option(mu_partial1_eqs)
    #mu_partial2_opt_eqs = sched_option(mu_partial2_eqs)
    #phi_partial1_opt_eqs = sched_option(phi_partial1_eqs)
    #phi_partial2_opt_eqs = sched_option(phi_partial2_eqs)

    mu_full_opt_kernel = ps.create_kernel(mu_full_opt_eqs, target='gpu', **optimization).compile()
    phi_full_opt_kernel = ps.create_kernel(phi_full_opt_eqs, target='gpu', **optimization).compile()

    #mu_partial1_opt_kernel = ps.create_staggered_kernel(mu_stag,
    #    mu_partial1_opt_eqs, target='gpu', **optimization).compile()
    #mu_partial2_opt_kernel = ps.create_kernel(
    #    mu_partial2_opt_eqs, target='gpu', **optimization).compile()

    #phi_partial1_opt_kernel = ps.create_kernel(
    #    phi_partial1_opt_eqs, target='gpu', **optimization).compile()
    #phi_partial2_opt_kernel = ps.create_kernel(
    #    phi_partial2_opt_eqs, target='gpu', **optimization).compile()

    bench_kernels(mu_full_opt_kernel, phi_kernel)
    #bench_kernels(mu_partial1_opt_kernel, phi_kernel)
    #bench_kernels(mu_partial2_opt_kernel, phi_kernel)

    bench_kernels(mu_kernel, phi_full_opt_kernel)
    #bench_kernels(mu_kernel, phi_partial1_opt_kernel)
    #bench_kernels(mu_kernel, phi_partial2_opt_kernel)

print()
